@@ -22,42 +22,39 @@
  *
  ***************************************************************************/
 
-#include "curl_setup.h"
+#include "../curl_setup.h"
 
 #ifdef USE_SSL
 
 #ifdef HAVE_SYS_TYPES_H
 #include <sys/types.h>
 #endif
-#ifdef HAVE_SYS_STAT_H
-#include <sys/stat.h>
-#endif
-#ifdef HAVE_FCNTL_H
-#include <fcntl.h>
-#endif
 
-#include "urldata.h"
-#include "cfilters.h"
+#include "../urldata.h"
+#include "../cfilters.h"
 
 #include "vtls.h" /* generic SSL protos etc */
 #include "vtls_int.h"
 #include "vtls_scache.h"
 #include "vtls_spack.h"
 
-#include "strcase.h"
-#include "url.h"
-#include "llist.h"
-#include "share.h"
-#include "curl_trc.h"
-#include "curl_sha256.h"
-#include "rand.h"
-#include "warnless.h"
-#include "curl_printf.h"
-#include "strdup.h"
+#include "../strcase.h"
+#include "../url.h"
+#include "../llist.h"
+#include "../share.h"
+#include "../curl_trc.h"
+#include "../curl_sha256.h"
+#include "../rand.h"
+#include "../curlx/warnless.h"
+#include "../curl_printf.h"
+#include "../strdup.h"
 
 /* The last #include files should be: */
-#include "curl_memory.h"
-#include "memdebug.h"
+#include "../curl_memory.h"
+#include "../memdebug.h"
+
+
+static bool cf_ssl_peer_key_is_global(const char *peer_key);
 
 /* a peer+tls-config we cache sessions for */
 struct Curl_ssl_scache_peer {
@@ -73,6 +70,7 @@ struct Curl_ssl_scache_peer {
   size_t max_sessions;
   long age;                /* just a number, the higher the more recent */
   BIT(hmac_set);           /* if key_salt and key_hmac are present */
+  BIT(exportable);         /* sessions for this peer can be exported */
 };
 
 #define CURL_SCACHE_MAGIC 0x000e1551
@@ -104,32 +102,18 @@ static struct Curl_ssl_scache *cf_ssl_scache_get(struct Curl_easy *data)
   return scache;
 }
 
-static void cf_ssl_scache_clear_session(struct Curl_ssl_session *s)
-{
-  if(s->sdata) {
-    free((void *)s->sdata);
-    s->sdata = NULL;
-  }
-  s->sdata_len = 0;
-  if(s->quic_tp) {
-    free((void *)s->quic_tp);
-    s->quic_tp = NULL;
-  }
-  s->quic_tp_len = 0;
-  s->ietf_tls_id = 0;
-  s->valid_until = 0;
-  Curl_safefree(s->alpn);
-}
-
-static void cf_ssl_scache_sesssion_ldestroy(void *udata, void *s)
+static void cf_ssl_scache_session_ldestroy(void *udata, void *obj)
 {
+  struct Curl_ssl_session *s = obj;
   (void)udata;
-  cf_ssl_scache_clear_session(s);
+  free(CURL_UNCONST(s->sdata));
+  free(CURL_UNCONST(s->quic_tp));
+  free((void *)s->alpn);
   free(s);
 }
 
 CURLcode
-Curl_ssl_session_create(unsigned char *sdata, size_t sdata_len,
+Curl_ssl_session_create(void *sdata, size_t sdata_len,
                         int ietf_tls_id, const char *alpn,
                         curl_off_t valid_until, size_t earlydata_max,
                         struct Curl_ssl_session **psession)
@@ -140,7 +124,7 @@ Curl_ssl_session_create(unsigned char *sdata, size_t sdata_len,
 }
 
 CURLcode
-Curl_ssl_session_create2(unsigned char *sdata, size_t sdata_len,
+Curl_ssl_session_create2(void *sdata, size_t sdata_len,
                          int ietf_tls_id, const char *alpn,
                          curl_off_t valid_until, size_t earlydata_max,
                          unsigned char *quic_tp, size_t quic_tp_len,
@@ -171,7 +155,7 @@ Curl_ssl_session_create2(unsigned char *sdata, size_t sdata_len,
   if(alpn) {
     s->alpn = strdup(alpn);
     if(!s->alpn) {
-      cf_ssl_scache_sesssion_ldestroy(NULL, s);
+      cf_ssl_scache_session_ldestroy(NULL, s);
       return CURLE_OUT_OF_MEMORY;
     }
   }
@@ -186,7 +170,7 @@ void Curl_ssl_session_destroy(struct Curl_ssl_session *s)
     if(Curl_node_llist(&s->list))
       Curl_node_remove(&s->list);
     else {
-      cf_ssl_scache_sesssion_ldestroy(NULL, s);
+      cf_ssl_scache_session_ldestroy(NULL, s);
     }
   }
 }
@@ -223,6 +207,19 @@ static void cf_ssl_scache_peer_set_obj(struct Curl_ssl_scache_peer *peer,
   peer->sobj_free = sobj_free;
 }
 
+static void cf_ssl_cache_peer_update(struct Curl_ssl_scache_peer *peer)
+{
+  /* The sessions of this peer are exportable if
+   * - it has no confidential information
+   * - its peer key is not yet known, because sessions were
+   *   imported using only the salt+hmac
+   * - the peer key is global, e.g. carrying no relative paths */
+  peer->exportable = (!peer->clientcert && !peer->srp_username &&
+                      !peer->srp_password &&
+                      (!peer->ssl_peer_key ||
+                       cf_ssl_peer_key_is_global(peer->ssl_peer_key)));
+}
+
 static CURLcode
 cf_ssl_scache_peer_init(struct Curl_ssl_scache_peer *peer,
                         const char *ssl_peer_key,
@@ -265,6 +262,8 @@ cf_ssl_scache_peer_init(struct Curl_ssl_scache_peer *peer,
     if(!peer->srp_password)
       goto out;
   }
+
+  cf_ssl_cache_peer_update(peer);
   result = CURLE_OK;
 out:
   if(result)
@@ -336,7 +335,7 @@ CURLcode Curl_ssl_scache_create(size_t max_peers,
   for(i = 0; i < scache->peer_count; ++i) {
     scache->peers[i].max_sessions = max_sessions_per_peer;
     Curl_llist_init(&scache->peers[i].sessions,
-                    cf_ssl_scache_sesssion_ldestroy);
+                    cf_ssl_scache_session_ldestroy);
   }
 
   *pscache = scache;
@@ -371,50 +370,56 @@ void Curl_ssl_scache_unlock(struct Curl_easy *data)
 }
 
 static CURLcode cf_ssl_peer_key_add_path(struct dynbuf *buf,
-                                          const char *name,
-                                          char *path)
+                                         const char *name,
+                                         char *path,
+                                         bool *is_local)
 {
   if(path && path[0]) {
     /* We try to add absolute paths, so that the session key can stay
      * valid when used in another process with different CWD. However,
      * when a path does not exist, this does not work. Then, we add
      * the path as is. */
-#ifdef _WIN32
+#ifdef UNDER_CE
+    (void)is_local;
+    return curlx_dyn_addf(buf, ":%s-%s", name, path);
+#elif defined(_WIN32)
     char abspath[_MAX_PATH];
     if(_fullpath(abspath, path, _MAX_PATH))
-      return Curl_dyn_addf(buf, ":%s-%s", name, abspath);
+      return curlx_dyn_addf(buf, ":%s-%s", name, abspath);
+    *is_local = TRUE;
 #elif defined(HAVE_REALPATH)
     if(path[0] != '/') {
       char *abspath = realpath(path, NULL);
       if(abspath) {
-        CURLcode r = Curl_dyn_addf(buf, ":%s-%s", name, abspath);
+        CURLcode r = curlx_dyn_addf(buf, ":%s-%s", name, abspath);
         (free)(abspath); /* allocated by libc, free without memdebug */
         return r;
       }
+      *is_local = TRUE;
     }
 #endif
-    return Curl_dyn_addf(buf, ":%s-%s", name, path);
+    return curlx_dyn_addf(buf, ":%s-%s", name, path);
   }
   return CURLE_OK;
 }
 
 static CURLcode cf_ssl_peer_key_add_hash(struct dynbuf *buf,
-                                          const char *name,
-                                          struct curl_blob *blob)
+                                         const char *name,
+                                         struct curl_blob *blob)
 {
   CURLcode r = CURLE_OK;
   if(blob && blob->len) {
     unsigned char hash[CURL_SHA256_DIGEST_LENGTH];
     size_t i;
 
-    r = Curl_dyn_addf(buf, ":%s-", name);
+    r = curlx_dyn_addf(buf, ":%s-", name);
     if(r)
       goto out;
     r = Curl_sha256it(hash, blob->data, blob->len);
     if(r)
       goto out;
     for(i = 0; i < CURL_SHA256_DIGEST_LENGTH; ++i) {
-      r = Curl_dyn_addf(buf, "%02x", hash[i]);
+      r = curlx_dyn_addf(buf, "%02x", hash[i]);
       if(r)
         goto out;
     }
@@ -423,6 +428,17 @@ out:
   return r;
 }
 
+#define CURL_SSLS_LOCAL_SUFFIX     ":L"
+#define CURL_SSLS_GLOBAL_SUFFIX    ":G"
+
+static bool cf_ssl_peer_key_is_global(const char *peer_key)
+{
+  size_t len = peer_key ? strlen(peer_key) : 0;
+  return (len > 2) &&
+         (peer_key[len - 1] == 'G') &&
+         (peer_key[len - 2] == ':');
+}
+
 CURLcode Curl_ssl_peer_key_make(struct Curl_cfilter *cf,
                                 const struct ssl_peer *peer,
                                 const char *tls_id,
@@ -431,12 +447,13 @@ CURLcode Curl_ssl_peer_key_make(struct Curl_cfilter *cf,
   struct ssl_primary_config *ssl = Curl_ssl_cf_get_primary_config(cf);
   struct dynbuf buf;
   size_t key_len;
+  bool is_local = FALSE;
   CURLcode r;
 
   *ppeer_key = NULL;
-  Curl_dyn_init(&buf, 10 * 1024);
+  curlx_dyn_init(&buf, 10 * 1024);
 
-  r = Curl_dyn_addf(&buf, "%s:%d", peer->hostname, peer->port);
+  r = curlx_dyn_addf(&buf, "%s:%d", peer->hostname, peer->port);
   if(r)
     goto out;
 
@@ -444,86 +461,86 @@ CURLcode Curl_ssl_peer_key_make(struct Curl_cfilter *cf,
   case TRNSPRT_TCP:
     break;
   case TRNSPRT_UDP:
-    r = Curl_dyn_add(&buf, ":UDP");
+    r = curlx_dyn_add(&buf, ":UDP");
     break;
   case TRNSPRT_QUIC:
-    r = Curl_dyn_add(&buf, ":QUIC");
+    r = curlx_dyn_add(&buf, ":QUIC");
     break;
   case TRNSPRT_UNIX:
-    r = Curl_dyn_add(&buf, ":UNIX");
+    r = curlx_dyn_add(&buf, ":UNIX");
     break;
   default:
-    r = Curl_dyn_addf(&buf, ":TRNSPRT-%d", peer->transport);
+    r = curlx_dyn_addf(&buf, ":TRNSPRT-%d", peer->transport);
     break;
   }
   if(r)
     goto out;
 
   if(!ssl->verifypeer) {
-    r = Curl_dyn_add(&buf, ":NO-VRFY-PEER");
+    r = curlx_dyn_add(&buf, ":NO-VRFY-PEER");
     if(r)
       goto out;
   }
   if(!ssl->verifyhost) {
-    r = Curl_dyn_add(&buf, ":NO-VRFY-HOST");
+    r = curlx_dyn_add(&buf, ":NO-VRFY-HOST");
     if(r)
       goto out;
   }
   if(ssl->verifystatus) {
-    r = Curl_dyn_add(&buf, ":VRFY-STATUS");
+    r = curlx_dyn_add(&buf, ":VRFY-STATUS");
     if(r)
       goto out;
   }
   if(!ssl->verifypeer || !ssl->verifyhost) {
     if(cf->conn->bits.conn_to_host) {
-      r = Curl_dyn_addf(&buf, ":CHOST-%s", cf->conn->conn_to_host.name);
+      r = curlx_dyn_addf(&buf, ":CHOST-%s", cf->conn->conn_to_host.name);
       if(r)
         goto out;
     }
     if(cf->conn->bits.conn_to_port) {
-      r = Curl_dyn_addf(&buf, ":CPORT-%d", cf->conn->conn_to_port);
+      r = curlx_dyn_addf(&buf, ":CPORT-%d", cf->conn->conn_to_port);
       if(r)
         goto out;
     }
   }
 
   if(ssl->version || ssl->version_max) {
-    r = Curl_dyn_addf(&buf, ":TLSVER-%d-%d", ssl->version,
+    r = curlx_dyn_addf(&buf, ":TLSVER-%d-%d", ssl->version,
                       (ssl->version_max >> 16));
     if(r)
       goto out;
   }
   if(ssl->ssl_options) {
-    r = Curl_dyn_addf(&buf, ":TLSOPT-%x", ssl->ssl_options);
+    r = curlx_dyn_addf(&buf, ":TLSOPT-%x", ssl->ssl_options);
     if(r)
       goto out;
   }
   if(ssl->cipher_list) {
-    r = Curl_dyn_addf(&buf, ":CIPHER-%s", ssl->cipher_list);
+    r = curlx_dyn_addf(&buf, ":CIPHER-%s", ssl->cipher_list);
     if(r)
       goto out;
   }
   if(ssl->cipher_list13) {
-    r = Curl_dyn_addf(&buf, ":CIPHER13-%s", ssl->cipher_list13);
+    r = curlx_dyn_addf(&buf, ":CIPHER13-%s", ssl->cipher_list13);
     if(r)
       goto out;
   }
   if(ssl->curves) {
-    r = Curl_dyn_addf(&buf, ":CURVES-%s", ssl->curves);
+    r = curlx_dyn_addf(&buf, ":CURVES-%s", ssl->curves);
     if(r)
       goto out;
   }
   if(ssl->verifypeer) {
-    r = cf_ssl_peer_key_add_path(&buf, "CA", ssl->CAfile);
+    r = cf_ssl_peer_key_add_path(&buf, "CA", ssl->CAfile, &is_local);
     if(r)
       goto out;
-    r = cf_ssl_peer_key_add_path(&buf, "CApath", ssl->CApath);
+    r = cf_ssl_peer_key_add_path(&buf, "CApath", ssl->CApath, &is_local);
     if(r)
       goto out;
-    r = cf_ssl_peer_key_add_path(&buf, "CRL", ssl->CRLfile);
+    r = cf_ssl_peer_key_add_path(&buf, "CRL", ssl->CRLfile, &is_local);
     if(r)
       goto out;
-    r = cf_ssl_peer_key_add_path(&buf, "Issuer", ssl->issuercert);
+    r = cf_ssl_peer_key_add_path(&buf, "Issuer", ssl->issuercert, &is_local);
     if(r)
       goto out;
     if(ssl->cert_blob) {
@@ -543,19 +560,19 @@ CURLcode Curl_ssl_peer_key_make(struct Curl_cfilter *cf,
     }
   }
   if(ssl->pinned_key && ssl->pinned_key[0]) {
-    r = Curl_dyn_addf(&buf, ":Pinned-%s", ssl->pinned_key);
+    r = curlx_dyn_addf(&buf, ":Pinned-%s", ssl->pinned_key);
     if(r)
       goto out;
   }
 
   if(ssl->clientcert && ssl->clientcert[0]) {
-    r = Curl_dyn_add(&buf, ":CCERT");
+    r = curlx_dyn_add(&buf, ":CCERT");
     if(r)
       goto out;
   }
 #ifdef USE_TLS_SRP
   if(ssl->username || ssl->password) {
-    r = Curl_dyn_add(&buf, ":SRP-AUTH");
+    r = curlx_dyn_add(&buf, ":SRP-AUTH");
     if(r)
       goto out;
   }
@@ -565,17 +582,21 @@ CURLcode Curl_ssl_peer_key_make(struct Curl_cfilter *cf,
     r = CURLE_FAILED_INIT;
     goto out;
   }
-  r = Curl_dyn_addf(&buf, ":IMPL-%s", tls_id);
+  r = curlx_dyn_addf(&buf, ":IMPL-%s", tls_id);
   if(r)
     goto out;
 
-  *ppeer_key = Curl_dyn_take(&buf, &key_len);
-  /* we just added printable char, and dynbuf always 0 terminates,
-   * no need to track length */
+  r = curlx_dyn_addf(&buf, is_local ?
+                     CURL_SSLS_LOCAL_SUFFIX : CURL_SSLS_GLOBAL_SUFFIX);
+  if(r)
+    goto out;
 
+  *ppeer_key = curlx_dyn_take(&buf, &key_len);
+  /* we just added printable char, and dynbuf always null-terminates, no need
+   * to track length */
 
 out:
-  Curl_dyn_free(&buf);
+  curlx_dyn_free(&buf);
   return r;
 }
 
@@ -622,7 +643,7 @@ cf_ssl_find_peer_by_key(struct Curl_easy *data,
   /* check for entries with known peer_key */
   for(i = 0; scache && i < scache->peer_count; i++) {
     if(scache->peers[i].ssl_peer_key &&
-       strcasecompare(ssl_peer_key, scache->peers[i].ssl_peer_key) &&
+       curl_strequal(ssl_peer_key, scache->peers[i].ssl_peer_key) &&
        cf_ssl_scache_match_auth(&scache->peers[i], conn_config)) {
       /* yes, we have a cached session for this! */
       *ppeer = &scache->peers[i];
@@ -655,6 +676,7 @@ cf_ssl_find_peer_by_key(struct Curl_easy *data,
           result = CURLE_OUT_OF_MEMORY;
           goto out;
         }
+        cf_ssl_cache_peer_update(&scache->peers[i]);
         *ppeer = &scache->peers[i];
         goto out;
       }
@@ -832,10 +854,6 @@ CURLcode Curl_ssl_scache_put(struct Curl_cfilter *cf,
     Curl_ssl_session_destroy(s);
     return CURLE_OK;
   }
-  if(!GOOD_SCACHE(scache)) {
-    Curl_ssl_session_destroy(s);
-    return CURLE_BAD_FUNCTION_ARGUMENT;
-  }
 
   Curl_ssl_scache_lock(data);
   result = cf_scache_add_session(cf, data, scache, ssl_peer_key, s);
@@ -844,9 +862,9 @@ CURLcode Curl_ssl_scache_put(struct Curl_cfilter *cf,
 }
 
 void Curl_ssl_scache_return(struct Curl_cfilter *cf,
-                           struct Curl_easy *data,
-                           const char *ssl_peer_key,
-                           struct Curl_ssl_session *s)
+                            struct Curl_easy *data,
+                            const char *ssl_peer_key,
+                            struct Curl_ssl_session *s)
 {
   /* See RFC 8446 C.4:
    * "Clients SHOULD NOT reuse a ticket for multiple connections." */
@@ -933,31 +951,29 @@ out:
   return result;
 }
 
-bool Curl_ssl_scache_get_obj(struct Curl_cfilter *cf,
-                             struct Curl_easy *data,
-                             const char *ssl_peer_key,
-                             void **sobj)
+void *Curl_ssl_scache_get_obj(struct Curl_cfilter *cf,
+                              struct Curl_easy *data,
+                              const char *ssl_peer_key)
 {
   struct Curl_ssl_scache *scache = cf_ssl_scache_get(data);
   struct ssl_primary_config *conn_config = Curl_ssl_cf_get_primary_config(cf);
   struct Curl_ssl_scache_peer *peer = NULL;
   CURLcode result;
+  void *sobj;
 
-  *sobj = NULL;
   if(!scache)
-    return FALSE;
+    return NULL;
 
   result = cf_ssl_find_peer_by_key(data, scache, ssl_peer_key, conn_config,
                                    &peer);
   if(result)
-    return FALSE;
+    return NULL;
 
-  if(peer)
-    *sobj = peer->sobj;
+  sobj = peer ? peer->sobj : NULL;
 
   CURL_TRC_SSLS(data, "%s cached session for '%s'",
-                *sobj ? "Found" : "No", ssl_peer_key);
-  return !!*sobj;
+                sobj ? "Found" : "No", ssl_peer_key);
+  return sobj;
 }
 
 void Curl_ssl_scache_remove_all(struct Curl_cfilter *cf,
@@ -1062,7 +1078,7 @@ out:
 CURLcode Curl_ssl_session_import(struct Curl_easy *data,
                                  const char *ssl_peer_key,
                                  const unsigned char *shmac, size_t shmac_len,
-                                 const unsigned char *sdata, size_t sdata_len)
+                                 const void *sdata, size_t sdata_len)
 {
   struct Curl_ssl_scache *scache = cf_ssl_scache_get(data);
   struct Curl_ssl_scache_peer *peer = NULL;
@@ -1150,17 +1166,17 @@ CURLcode Curl_ssl_session_export(struct Curl_easy *data,
 
   Curl_ssl_scache_lock(data);
 
-  Curl_dyn_init(&hbuf, (CURL_SHA256_DIGEST_LENGTH * 2) + 1);
-  Curl_dyn_init(&sbuf, CURL_SSL_TICKET_MAX);
+  curlx_dyn_init(&hbuf, (CURL_SHA256_DIGEST_LENGTH * 2) + 1);
+  curlx_dyn_init(&sbuf, CURL_SSL_TICKET_MAX);
 
   for(i = 0; scache && i < scache->peer_count; i++) {
     peer = &scache->peers[i];
     if(!peer->ssl_peer_key && !peer->hmac_set)
       continue;  /* skip free entry */
-    if(peer->clientcert || peer->srp_username || peer->srp_password)
-      continue;  /* not exporting those */
+    if(!peer->exportable)
+      continue;
 
-    Curl_dyn_reset(&hbuf);
+    curlx_dyn_reset(&hbuf);
     cf_scache_peer_remove_expired(peer, now);
     n = Curl_llist_head(&peer->sessions);
     if(n)
@@ -1172,22 +1188,22 @@ CURLcode Curl_ssl_session_export(struct Curl_easy *data,
         if(r)
           goto out;
       }
-      if(!Curl_dyn_len(&hbuf)) {
-        r = Curl_dyn_addn(&hbuf, peer->key_salt, sizeof(peer->key_salt));
+      if(!curlx_dyn_len(&hbuf)) {
+        r = curlx_dyn_addn(&hbuf, peer->key_salt, sizeof(peer->key_salt));
         if(r)
           goto out;
-        r = Curl_dyn_addn(&hbuf, peer->key_hmac, sizeof(peer->key_hmac));
+        r = curlx_dyn_addn(&hbuf, peer->key_hmac, sizeof(peer->key_hmac));
         if(r)
           goto out;
       }
-      Curl_dyn_reset(&sbuf);
+      curlx_dyn_reset(&sbuf);
       r = Curl_ssl_session_pack(data, s, &sbuf);
       if(r)
         goto out;
 
       r = export_fn(data, userptr, peer->ssl_peer_key,
-                    Curl_dyn_uptr(&hbuf), Curl_dyn_len(&hbuf),
-                    Curl_dyn_uptr(&sbuf), Curl_dyn_len(&sbuf),
+                    curlx_dyn_uptr(&hbuf), curlx_dyn_len(&hbuf),
+                    curlx_dyn_uptr(&sbuf), curlx_dyn_len(&sbuf),
                     s->valid_until, s->ietf_tls_id,
                     s->alpn, s->earlydata_max);
       if(r)
@@ -1203,8 +1219,8 @@ CURLcode Curl_ssl_session_export(struct Curl_easy *data,
 
 out:
   Curl_ssl_scache_unlock(data);
-  Curl_dyn_free(&hbuf);
-  Curl_dyn_free(&sbuf);
+  curlx_dyn_free(&hbuf);
+  curlx_dyn_free(&sbuf);
   return r;
 }
 
