@@ -24,20 +24,17 @@
 #include "tool_setup.h"
 
 #ifdef HAVE_PWD_H
-#  undef __NO_NET_API /* required for building for AmigaOS */
-#  include <pwd.h>
+#ifdef __AMIGA__
+#undef __NO_NET_API /* required for AmigaOS to declare getpwuid() */
 #endif
-
-#ifdef HAVE_SYS_STAT_H
-#include <sys/stat.h>
+#include <pwd.h>
+#ifdef __AMIGA__
+#define __NO_NET_API
 #endif
-#ifdef HAVE_FCNTL_H
-#include <fcntl.h>
 #endif
 
-#include <curlx.h>
-
 #include "tool_findfile.h"
+#include "tool_cfgable.h"
 
 #include "memdebug.h" /* keep this as LAST include */
 
@@ -76,7 +73,7 @@ static char *checkhome(const char *home, const char *fname, bool dotscore)
     else
       c = aprintf("%s" DIR_CHAR "%s", home, fname);
     if(c) {
-      int fd = open(c, O_RDONLY);
+      int fd = curlx_open(c, O_RDONLY);
       if(fd >= 0) {
         char *path = strdup(c);
         close(fd);
