@@ -51,11 +51,9 @@ static void t518_store_errmsg(const char *msg, int err)
 {
   if(!err)
     curl_msnprintf(t518_msgbuff, sizeof(t518_msgbuff), "%s", msg);
-  else {
-    char errbuf[STRERROR_LEN];
+  else
     curl_msnprintf(t518_msgbuff, sizeof(t518_msgbuff), "%s, errno %d, %s", msg,
-                   err, curlx_strerror(err, errbuf, sizeof(errbuf)));
-  }
+                   err, strerror(err));
 }
 
 static void t518_close_file_descriptors(void)
